/* zeroer - Clears unused disk space                            */
/* Revision: mod-110214                                         */
/* License:  GNU General Public License                         */
/* Credits:  Based on a program by David Schneider              */

/* ------------------------------------------------------------ */
/*                  set up large-file support                   */
/* ------------------------------------------------------------ */

#define _FILE_OFFSET_BITS      64
#define _LARGEFILE_SOURCE       1
#define _LARGEFILE64_SOURCE     1

/* ------------------------------------------------------------ */
/*                    standard header files                     */
/* ------------------------------------------------------------ */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <popt.h>
#include <sys/stat.h>
#include <sys/types.h>

/* ------------------------------------------------------------ */
/*                       basic constants                        */
/* ------------------------------------------------------------ */

#define ZERO    0
#define ONE     1
#define TWO     2
#define BYTES_PER_MEGABYTE  ((off_t) 1048576L)

/* ------------------------------------------------------------ */
/*                      program parameters                      */
/* ------------------------------------------------------------ */

                                /* Default block size (bytes)   */
#define DEFAULT_BLOCK_SIZE  ((off_t) (10 * BYTES_PER_MEGABYTE))

#define PATHBUFSIZE         128 /* Pathname buffer size to use  */

/* ------------------------------------------------------------ */
/*                       global variables                       */
/* ------------------------------------------------------------ */

char *TempFileFinalPath = NULL;
FILE *TempFileOFP       = NULL;
off_t zeblocksize       = DEFAULT_BLOCK_SIZE;
poptContext optCtx;

/* ------------------------------------------------------------ */

/* Note:  For  information about  this structure,  execute "man */
/* popt".                                                       */

struct poptOption commandOptions[] =
{
    { "blocksize" , 'b' , POPT_ARG_LONG , &zeblocksize ,
                    'b' , NULL          , NULL       } ,

    POPT_AUTOHELP               /* For -?, --help, --usage      */
    POPT_TABLEEND               /* Marks end of structure       */
};

/* ------------------------------------------------------------ */
/*                       support routines                       */
/* ------------------------------------------------------------ */

void ProgExit (int code)
{
    poptFreeContext (optCtx);
    exit (code);
}

/* ------------------------------------------------------------ */

void errExit (char *msg)
{
    perror (msg);
    ProgExit (EXIT_FAILURE);
}

/* ------------------------------------------------------------ */

void SigHandler (void)
{
    fprintf (stderr, "\nReceived a shutdown signal\n");
    fflush (stderr);

    if (TempFileOFP != NULL)
    {
        fclose (TempFileOFP);
        TempFileOFP = NULL;
    }

    if (TempFileFinalPath != NULL)
    {
#ifdef NOTDEF
        unlink (TempFileFinalPath);
        fprintf (stderr, "Removed temporary file: %s\n",
            TempFileFinalPath);
        fflush (stderr);
#endif
        TempFileFinalPath = NULL;
    }

    ProgExit (EXIT_FAILURE);
}

/* ------------------------------------------------------------ */

void blockWriteLoop (FILE *ofp, char *pBuf, char *tmpfilename)
{
                                /* Current output-block size    */
    off_t WriteSize    = zeblocksize;
    off_t BytesWritten = ZERO;  /* Total no. of bytes written   */

    struct stat sbuf;           /* "stat"-information buffer    */
    char *cp;                   /* Scratch                      */

    while (WriteSize > ZERO)
    {
        if (BytesWritten > ZERO)
        {
            off_t mb = BytesWritten / BYTES_PER_MEGABYTE;
            cp = "bytes";
            if (BytesWritten == ONE) cp = "byte ";
            printf ("%8lld MB %13lld %s written\n",
                (long long) mb, (long long) BytesWritten, cp);
            fflush (stdout);
        }

        fwrite ((void *) pBuf,
                (size_t) sizeof (char), WriteSize, ofp);
        BytesWritten += WriteSize;

        if (fflush (ofp) || ferror (ofp))
        {
            clearerr (ofp);
            WriteSize = WriteSize / TWO;

            if (WriteSize > ZERO)
            {
                printf ("\nreducing block size to %lld\n",
                    (long long) WriteSize);
                fflush (stdout);
            }
                                /* Get updated "stat" info      */
            if (stat (tmpfilename, &sbuf) != ZERO)
            {                   /* Error                        */
                fprintf (stderr,
                    "Undetermined error: Lost temporary file\n");
                fclose (ofp);
                                /* Just in case                 */
                unlink (tmpfilename);
                ProgExit (EXIT_FAILURE);
            }
                                /* Current temporary-file size  */
            BytesWritten = sbuf.st_size;
        }
    }

    cp = "bytes";
    if (BytesWritten == ONE) cp = "byte";
    printf ("\nWrote %lld %s total to %s\n",
        (long long) BytesWritten, cp, tmpfilename);
    fflush (stdout);
}

/* ------------------------------------------------------------ */

void ProcTarget (char *tmpfilename)
{
    FILE *ofp  = NULL;          /* Output-stream pointer        */
    char *pBuf = NULL;          /* Output-block  pointer        */

    int n;                      /* Scratch                      */
    struct stat sbuf;           /* "stat"-information buffer    */
    char xname [PATHBUFSIZE];   /* Temporary-file name buffer   */

/* ------------------------------------------------------------ */

/* Handle a directory argument  */

    if ((stat (tmpfilename, &sbuf) == ZERO) &&
        ((sbuf.st_mode & S_IFDIR) != ZERO))
    {                           /* Specified a directory        */
        if ((n = strlen (tmpfilename)) >= (PATHBUFSIZE-30))
        {
            fprintf (stderr, "%s:\n%s\n" ,
"Error: Pathname is too long" ,
tmpfilename);
            ProgExit (EXIT_FAILURE);
        }
                                /* Copy directory path          */
        strcpy (xname, tmpfilename);
                                /* Add a slash if necessary     */
        if (xname [n-ONE] != '/') strcat (xname, "/");

                                /* Add a temporary-file name    */
        strcat (xname, "zeroer-temp-data");
        unlink (xname);         /* Remove existing copy, if any */

        tmpfilename = xname;    /* Point to filename instead of */
                                /* directory name               */

                                /* Does the file exist?         */
        if (stat (tmpfilename, &sbuf) == ZERO)
        {                       /* Yes - "unlink" failed        */
            fprintf (stderr, "%s:\n%s\n" ,
"Error: Couldn't delete temporary file" ,
tmpfilename);
            ProgExit (EXIT_FAILURE);
        }
                                /* Status message               */
        printf ("%s:\n%s\n",
"Specified a directory. Setting temporary-file pathname to" ,
tmpfilename);
    }

/* ------------------------------------------------------------ */

    if (stat (tmpfilename, &sbuf) == ZERO)
    {
        fprintf (stderr, "Error: %s:\n%s\n" ,
"Specified file already exists [or it's a link]" ,
tmpfilename);
        fprintf (stderr, "\n%s\n%s\n%s\n%s\n",
"You can specify the pathname for a directory, which must exist. Alter-" ,
"natively,  you can specify the pathname  for a file,  which must _not_" ,
"exist. Either way, this program will clear all free space in the asso-" ,
"ciated filesystem.");

    fprintf (stderr, "\n%s\n",
"For more information, execute \"man zeroer\".");

        ProgExit (EXIT_FAILURE);
    }

/* ------------------------------------------------------------ */

    if ((ofp = fopen (tmpfilename, "wb")) == NULL)
    {
        errExit ("\nCould not create temporary file");
    }

    TempFileFinalPath = tmpfilename;
    TempFileOFP       = ofp;
    signal (SIGINT  , (void *) SigHandler);
    signal (SIGTERM , (void *) SigHandler);

    pBuf = (char *) malloc ((size_t) (sizeof (char) * zeblocksize));

    if (pBuf == NULL)
    {
        errExit ("\nCould not allocate dynamic memory");
    }

    memset ((void *) pBuf, ZERO,
            (size_t) (sizeof(char) * zeblocksize));
    blockWriteLoop (ofp, pBuf, tmpfilename);

    free ((void *) pBuf);
    fclose (ofp);
    sleep (TWO);

#ifdef NOTDEF
    unlink (tmpfilename);
                                /* Status messages              */
    printf ("Removed temporary file: %s\n", TempFileFinalPath);
#endif

    puts ("Done");
    fflush (stdout);
}

/* ------------------------------------------------------------ */

static void UsageError()
{
    fprintf (stderr, "%s\n%s\n",
"zeroer mod-061203 - Clears unused disk space" ,
"Based on zeroer 0.1 by David Schneider");

    fprintf (stderr, "\n%s\n%s\n%s\n",
"Usage: zeroer DIRECTORY" ,
"or     zeroer FILENAME"  ,
"or     zeroer -b BLOCKSIZE DIRECTORY_or_FILENAME");

    fprintf (stderr, "\n%s\n%s\n%s\n%s\n",
"You can specify the pathname for a directory, which must exist. Alter-" ,
"natively,  you can specify the  pathname for a file,  which must _not_" ,
"exist.  Either way, this program will  clear all free space in the as-" ,
"sociated filesystem.");

    fprintf (stderr, "\n%s\n",
"For more information, execute \"man zeroer\".");

    ProgExit (EXIT_FAILURE);
}

/* ------------------------------------------------------------ */
/*                         main routine                         */
/* ------------------------------------------------------------ */

int main (int argc, char **argv)
{
    int opt;
    char *tmpfilename;          /* Target-object pathname       */

    optCtx = poptGetContext     /* Initialize "popt" subsystem  */
    (
        NULL, argc, (const char **) argv, commandOptions, ZERO
    );

    poptSetOtherOptionHelp (optCtx, "filename");
    if (argc < TWO) UsageError();
                                /* Parse option switches        */
    while ((opt = poptGetNextOpt (optCtx)) >= ZERO)
    {
        switch (opt)
        {
    case 'b':
            fprintf (stderr, "blocksize set to %ld\n",
                zeblocksize);
            break;
        }
    }
                                /* Get target-object parameter  */
    tmpfilename = (char *) poptGetArg (optCtx);
    if (tmpfilename == NULL) UsageError();
    if (opt < -1) UsageError();

    ProcTarget (tmpfilename);   /* Process the specified object */
    ProgExit (EXIT_SUCCESS);
}
